// Copyright (C) 2019  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// XXX #define  _GNU_SOURCE ?

#include "wcfs_misc.h"

#include <golang/libgolang.h>
using namespace golang;

#include <inttypes.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>

#include <memory>


// io::
namespace io {

const error EOF_                = fmt::errorf("EOF");
const error ErrUnexpectedEOF    = fmt::errorf("unexpected EOF");

}   // io::

// os::
namespace os {

static error _pathError(const char *op, const string &path, int syserr);

int     File::fd() const    { return _fd; }
string  File::name() const  { return _path; }


tuple<File, error> open(const string &path, int flags, mode_t mode) {
    File f;
    f._path  = path;
    f._fd = ::open(path.c_str(), flags, mode);
    error err;
    if (f._fd == -1)
        err = f._errno("open");
    return make_tuple(f, err);
}

error File::close() {
    File &f = *this;

    int err = ::close(f._fd);
    if (err != 0)
        return f._errno("close");
    f._fd = -1;
    return nil;
}

tuple<int, error> File::read(void *buf, size_t count) {
    File &f = *this;
    int n;

    n = ::read(f._fd, buf, count);
    if (n == 0)
        return make_tuple(n, io::EOF_);
    if (n < 0)
        return make_tuple(0, f._errno("read"));

    return make_tuple(n, nil);
}

tuple <int, error> File::write(const void *buf, size_t count) {
    File &f = *this;
    int n, wrote=0;

    // NOTE contrary to write(2) we have to write all data as io.Writer requires.
    while (count != 0) {
        n = ::write(f._fd, buf, count);
        if (n < 0)
            return make_tuple(wrote, f._errno("write"));

        wrote += n;
        buf    = ((const char *)buf) + n;
        count -= n;
    }

    return make_tuple(wrote, nil);
}

error File::stat(struct stat *st) {
    File &f = *this;

    int err = fstat(f._fd, st);
    if (err != 0)
        return f._errno("stat");
    return nil;
}


// _errno returns error corresponding to op(file) and errno.
error File::_errno(const char *op) {
    File &f = *this;
    return _pathError(op, f._path, errno);
}

// _pathError returns os.PathError-like for op/path and system error
// indicated by syserr.
static error _pathError(const char *op, const string &path, int syserr) {
    char ebuf[128];
    char *estr = strerror_r(syserr, ebuf, sizeof(ebuf));
    return fmt::errorf("%s %s: %s", op, path.c_str(), estr);
}

}   // os::


// mm::
namespace mm {

// map_into memory-maps f.fd[offset +size) into [addr +size).
// prot  is PROT_* from mmap(2).
// flags is MAP_*  from mmap(2); MAP_FIXED is added automatically.
error map_into(void *addr, size_t size, int prot, int flags, const os::File &f, off_t offset) {
    void *addr2;

    addr2 = mmap(addr, size, prot, MAP_FIXED | flags, f.fd(), offset);
    if (addr2 == MAP_FAILED)
        return os::_pathError("mmap", f.name(), errno);
    if (addr2 != addr)
        panic("mmap(addr, MAP_FIXED): returned !addr");
    return nil;
}

}   // mm::


// fmt::
namespace fmt {

static
string _vsprintf(const char *format, va_list argp) {
    // based on https://stackoverflow.com/a/26221725/9456786
    va_list argp2;
    va_copy(argp2, argp);
    size_t size = vsnprintf(NULL, 0, format, argp2);
    va_end(argp2);

    std::unique_ptr<char[]> buf( new char[size] );
    vsnprintf(buf.get(), size, format, argp);
    return string(buf.get(), buf.get() + size - 1 ); // without trailing '\0'
}

string sprintf(const string &format, ...) {
    va_list argp;
    va_start(argp, format);
    string str = fmt::_vsprintf(format.c_str(), argp);
    va_end(argp);
    return str;
}

string sprintf(const char *format, ...) {
    va_list argp;
    va_start(argp, format);
    string str = fmt::_vsprintf(format, argp);
    va_end(argp);
    return str;
}

error errorf(const string &format, ...) {
    error err;
    va_list argp;
    va_start(argp, format);
    err.err = fmt::sprintf(format.c_str(), argp);
    va_end(argp);
    return err;
}

error errorf(const char *format, ...) {
    error err;
    va_list argp;
    va_start(argp, format);
    err.err = fmt::sprintf(format, argp);
    va_end(argp);
    return err;
}

}   // fmt::


// strings::
namespace strings {

bool has_prefix(const string &s, const string &prefix) {
    return s.compare(0, prefix.size(), prefix);
}

bool has_prefix(const string &s, char prefix) {
    return (s.size() >= 1 && s[0] == prefix);
}

vector<string> split(const string &s, char sep) {
    vector<string> r;
    int psep_prev=-1;
    size_t psep;
    while (1) {
        psep = s.find(sep, psep_prev+1);
        if (psep == string::npos)
            return r;

        r.push_back(s.substr(psep_prev+1, sep-(psep_prev+1)));
        psep_prev = sep;
    }
}

}   // strings::


// xstrconv::   (strconv-like)
namespace xstrconv {

// parseHex64 decodes 16-character-wide hex-encoded string into uint64.
tuple<uint64_t, error> parseHex64(const string& s) {
    if (s.size() != 16)
        return make_tuple(0, fmt::errorf("hex64 %s invalid", s.c_str()));

    uint64_t v;
    int n = sscanf(s.c_str(), "%16" SCNx64, &v);    // XXX verify
    if (n != 1)
        return make_tuple(0, fmt::errorf("hex64 %s invalid", s.c_str()));

    return make_tuple(v, nil);
}

// paeseInt decodes string s as signed decimal integer.
tuple<int64_t, error> parseInt(const string& s) {
    int64_t v;
    int n = sscanf(s.c_str(), "%" SCNi64, &v);      // XXX verify
    if (!(n == 1 && std::to_string(v) == s))        // XXX verify
        return make_tuple(0, fmt::errorf("int %s invalid", s.c_str()));
    return make_tuple(v, nil);
}

// parseUint decodes string s as unsigned decimal integer.
tuple<uint64_t, error> parseUint(const string& s) {
    uint64_t v;
    int n = sscanf(s.c_str(), "%" SCNu64, &v);      // XXX verify
    if (!(n == 1 && std::to_string(v) == s))        // XXX verify
        return make_tuple(0, fmt::errorf("uint %s invalid", s.c_str()));
    return make_tuple(v, nil);
}

}   // xstrconv::
