// Copyright (C) 2019  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// wcfs_misc.{h,cpp} provide miscellaneous utilities for other wcfs_* files.

#ifndef _NXD_WCFS_MISC_H_
#define _NXD_WCFS_MISC_H_

#include <stddef.h>
#include <stdint.h>

#include <string>
using std::string;

#include <tuple>
using std::tuple;
using std::make_tuple;
using std::tie;

#include <vector>
using std::vector;

// nil is synonym for nullptr and NULL.
const nullptr_t nil = nullptr;

// error mimics error from Go.
struct error {
    string err;	// XXX -> private + ,error() ?

    error()             {}
    error(nullptr_t)    {}  //  = nil

    // == nil
    bool operator==(nullptr_t) const {
        return err.empty();
    }
    bool operator!=(nullptr_t) const {
        return !(*this==NULL);
    }

    // == error
    bool operator==(const error &rhs) const {
        return (err == rhs.err);
    }
    bool operator!=(const error &rhs) const {
        return !(*this==rhs);
    }
};

// io::
namespace io {

extern const error EOF_;
extern const error ErrUnexpectedEOF;


}   // io::


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

// os::
namespace os {

// os::File mimics os.File from Go.
// its operations return error with full file context.
class File {
    int    _fd;
    string _path;

public:
    friend tuple<File, error> open(const string &path, int flags, mode_t mode);
    // XXX empty ctor -> fd=-1, path=?
    int     fd()    const;
    string  name()  const;
    error   close();

    // read implements io.Reader from Go: it reads into buf up-to count bytes.
    //
    // XXX buf -> slice<byte> ?
    tuple<int, error> read(void *buf, size_t count);

    // write implements io.Writer from Go: it writes all data from buf.
    //
    // NOTE write behaves like io.Writer in Go - it tries to write as much
    // bytes as requested, and if it could write only less - it returns error.
    // XXX buf -> slice<byte> ?
    tuple<int, error> write(const void *buf, size_t count);

    error   stat(struct stat *st);

private:
    error _errno(const char *op);
};

// open opens file @path.
tuple<File, error> open(const string &path, int flags = O_RDONLY,
        mode_t mode = S_IRUSR | S_IWUSR | S_IXUSR |
                      S_IRGRP | S_IWGRP | S_IXGRP |
                      S_IROTH | S_IWOTH | S_IXOTH);

}   // os::

// mm::
namespace mm {
    error map_into(void *addr, size_t, int prot, int flags, const os::File &f, off_t offset);
    // XXX unmap

}   // mm::


// fmt::
namespace fmt {

string sprintf(const string &format, ...);
error  errorf (const string &format, ...);

// `const char *` overload just to catch format mistakes as
// __attribute__(format) does not work with std::string.
string sprintf(const char *format, ...)
    __attribute__ ((format (printf, 1, 2)));
error  errorf (const char *format, ...)
    __attribute__ ((format (printf, 1, 2)));

}   // fmt::


//  strings::
namespace strings {

bool has_prefix(const string &s, const string &prefix);
bool has_prefix(const string &s, char prefix);
vector<string> split(const string &s, char sep);

}   // strings::


// XXX ok?
struct IContext {
    virtual chan<structZ>   done() = 0;
    virtual error           err() = 0;
};

// context::
namespace context {

struct _Background : IContext {
    chan<structZ>  done() {
        return nil;
    }

    error   err() {
        return nil;
    }
};

static _Background _bg;

// XXX doc
IContext* background() {
    return &_bg;    // NOTE nil is not valid in C++ (IContext* also carries vtab ptr)
}

// XXX doc
const error canceled = fmt::errorf("context canceled");
// XXX deadline exceeded?

}   // context::

#if 0
interface(Context) {
    ifunc(chan<structZ> done());
    ifunc(error         err());
};

I<io::Reader>(f)

// XXX wrap T* as IContext
template<typename T>
class Context : public IContext {
    T *obj;
public:
    Context(T *obj) : obj(obj) {}

    chan<structZ>   done()  { return obj->done();   }
    error           err()   { return obj->err();    }
};
#endif


// ---- misc ----

#include <unordered_map>
#include <unordered_set>

// dict wraps unordered_map into ergonomic interface.
template<typename Key, typename Value>
struct dict : std::unordered_map<Key, Value> {
    // has returns whether dict has element k.
    bool has(const Key &k) const {
        const dict &d = *this;
        return d.find(k) != d.end();
    }

    // get implements `d[k] -> (v, ok)`.
    tuple<Value, bool> get(const Key &k) const {
        const dict &d = *this;
        auto _ = d.find(k);
        if (_ == d.end())
            return make_tuple(Value(), false);
        return make_tuple(_->second, true);
    }

    // pop implements `d[k] -> (v, ok); del d[k]`.
    tuple<Value, bool> pop(const Key &k) {
        dict &d = *this;
        auto _ = d.find(k);
        if (_ == d.end())
            return make_tuple(Value(), false);
        Value v = _->second;
        d.erase(_);
        return make_tuple(v, true);
    }
};

// set wraps unordered_set into ergonomic interface.
template<typename Key>
struct set : std::unordered_set<Key> {
    // has returns whether set has element k.
    bool has(const Key &k) const {
        const set &s = *this;
        return s.find(k) != s.end();
    }
};


// xstrconv::
namespace xstrconv {

tuple<uint64_t, error> parseHex64(const string& s);
tuple<int64_t,  error> parseInt(const string& s);
tuple<uint64_t, error> parseUint(const string& s);

}   // xstrconv::


// zodb::
namespace zodb {

typedef uint64_t Tid;
typedef uint64_t Oid;

}   // zodb::

// TidHead is invalid Tid which is largest Tid value and means @head.
const zodb::Tid TidHead = -1ULL;

#endif
